#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "3dtypes.h"

extern object obj[];
extern int avertices,afaces,maxobj;

#ifndef PI
#define PI 3.141592687
#endif

typedef unsigned char  byte;
typedef unsigned short word;
typedef unsigned long  dword;

typedef signed char  sbyte;
typedef signed short sword;
typedef signed long  sdword;

#pragma pack(2)

typedef struct {
    word    id;
    dword   len;
} TChunkHeader, *PChunkHeader;

#pragma pack()

int obj_n; //numer objektu

enum {
    CHUNK_RGBF      = 0x0010,
    CHUNK_RGBB      = 0x0011,

    CHUNK_PRJ       = 0xC23D,
    CHUNK_MLI       = 0x3DAA,

    CHUNK_MAIN      = 0x4D4D,
        CHUNK_OBJMESH   = 0x3D3D,
            CHUNK_BKGCOLOR  = 0x1200,
            CHUNK_AMBCOLOR  = 0x2100,
            CHUNK_OBJBLOCK  = 0x4000,
                CHUNK_TRIMESH   = 0x4100,
                    CHUNK_VERTLIST  = 0x4110,
                    CHUNK_FACELIST  = 0x4120,
                    CHUNK_FACEMAT   = 0x4130,
                    CHUNK_MAPLIST   = 0x4140,
                    CHUNK_SMOOLIST  = 0x4150,
                    CHUNK_TRMATRIX  = 0x4160,
                CHUNK_LIGHT     = 0x4600,
                    CHUNK_SPOTLIGHT = 0x4610,
                CHUNK_CAMERA    = 0x4700,
                CHUNK_HIERARCHY = 0x4F00,
        CHUNK_VIEWPORT  = 0x7001,
        CHUNK_MATERIAL  = 0xAFFF,
            CHUNK_MATNAME   = 0xA000,
            CHUNK_AMBIENT   = 0xA010,
            CHUNK_DIFFUSE   = 0xA020,
            CHUNK_SPECULAR  = 0xA030,
            CHUNK_TEXTURE   = 0xA200,
            CHUNK_BUMPMAP   = 0xA230,
            CHUNK_MAPFILE   = 0xA300,
        CHUNK_KEYFRAMER = 0xB000,
            CHUNK_AMBIENTKEY    = 0xB001,
            CHUNK_TRACKINFO = 0xB002,
                CHUNK_TRACKOBJNAME  = 0xB010,
                CHUNK_TRACKPIVOT    = 0xB013,
                CHUNK_TRACKPOS      = 0xB020,
                CHUNK_TRACKROTATE   = 0xB021,
                CHUNK_TRACKSCALE    = 0xB022,
                CHUNK_OBJNUMBER     = 0xB030,
            CHUNK_TRACKCAMERA = 0xB003,
                CHUNK_TRACKFOV  = 0xB023,
                CHUNK_TRACKROLL = 0xB024,
            CHUNK_TRACKCAMTGT = 0xB004,
            CHUNK_TRACKLIGHT  = 0xB005,
            CHUNK_TRACKLIGTGT = 0xB006,
            CHUNK_TRACKSPOTL  = 0xB007,
            CHUNK_FRAMES    = 0xB008,


};

// ------------------------------------

    // Forward declaration.
void ChunkReader(FILE *f, int ind, long p);

void SkipReader(FILE *f, int ind, long p) {
}

void RGBFReader (FILE *f, int ind, long p) {
    float c[3];
    if (fread(&c, sizeof(c), 1, f) != 1) return;
}

void RGBBReader (FILE *f, int ind, long p) {
    byte c[3];
    if (fread(&c, sizeof(c), 1, f) != 1) return;
}

void ASCIIZReader (FILE *f, int ind, long p) {
    int c;

    while ( (c = fgetc(f)) != EOF && c != '\0');
//        putchar(c);
}

void ObjBlockReader (FILE *f, int ind, long p) {
    int c;

    ASCIIZReader(f, ind, p);
    ChunkReader(f, ind, p);
}

void VertListReader (FILE *f, int ind, long p) {
    word nv;
    float c[3];
    int i;
    v3d *v;

    if (fread(&nv, sizeof(nv), 1, f) != 1) return;

    obj[obj_n].v   = calloc(nv,sizeof(v3d));
    obj[obj_n].n2v = calloc(nv,sizeof(v3d));
    obj[obj_n].uv = calloc(nv,sizeof(v3d));
    obj[obj_n].sc  = calloc(nv,sizeof(p2d));
    obj[obj_n].vertices = nv;
    obj[obj_n].st = 0;
    obj[obj_n].p = 0;
    obj[obj_n].t = 0;
    avertices+=nv;

//    if (fread(obj[obj_n].v, sizeof(v3d), nv, f) != 1) return;
    v=obj[obj_n].v;
    for (i=0;i<nv;i++)
      {
        fread(v,3*sizeof(float),1,f);
        v++;
      }


}

void FaceListReader (FILE *f, int ind, long p) {
    word nv;
    word c[3];
    word flags;
    word i;

    if (fread(&nv, sizeof(nv), 1, f) != 1) return;

    obj[obj_n].n2f = calloc(nv,sizeof(v3d));
    obj[obj_n].wr  = calloc(nv,sizeof(f3d));
    obj[obj_n].faces = nv;

    afaces+=nv;

    i=0;

    while (nv-- > 0) {
        if (fread(&c, sizeof(c), 1, f) != 1) return;
        if (fread(&flags, sizeof(flags), 1, f) != 1) return;
        obj[obj_n].wr[i].a=c[0];
        obj[obj_n].wr[i].b=c[1];
        obj[obj_n].wr[i].c=c[2];
        i++;
    }


    obj_n++;
    maxobj=obj_n;
    ChunkReader(f, ind, p);
}

void FaceMatReader (FILE *f, int ind, long p) {
    int c;
    word n, nf;

    ASCIIZReader(f, ind, p);

    if (fread(&n, sizeof(n), 1, f) != 1) return;
    while (n-- > 0) {
        if (fread(&nf, sizeof(nf), 1, f) != 1) return;
    }
}

void MapListReader (FILE *f, int ind, long p) {
    word nv;
    float c[2];

    if (fread(&nv, sizeof(nv), 1, f) != 1) return;
    while (nv-- > 0) {
        if (fread(&c, sizeof(c), 1, f) != 1) return;
    }
}

void SmooListReader (FILE *f, int ind, long p) {
    dword s;
    int i;

    while (ftell(f) < p) {
        if (fread(&s, sizeof(s), 1, f) != 1) return;
    }
}

void TrMatrixReader(FILE *f, int ind, long p) {
    float rot[9];
    float trans[3];
    if (fread(&rot, sizeof(rot), 1, f) != 1) return;
    if (fread(&trans, sizeof(trans), 1, f) != 1) return;
}

void LightReader(FILE *f, int ind, long p) {
    float c[3];
    if (fread(&c, sizeof(c), 1, f) != 1) return;
    ChunkReader(f, ind, p);
}

void SpotLightReader(FILE *f, int ind, long p) {
    float c[5];
    if (fread(&c, sizeof(c), 1, f) != 1) return;
}
 
void CameraReader(FILE *f, int ind, long p) {
    float c[8];
    if (fread(&c, sizeof(c), 1, f) != 1) return;
}

void MatNameReader (FILE *f, int ind, long p) {
    int c;

    ASCIIZReader(f, ind, p);
}

void MapFileReader(FILE *f, int ind, long p) {
    int c;

    ASCIIZReader(f, ind, p);
}

void FramesReader(FILE *f, int ind, long p) {
    dword c[2];
    if (fread(&c, sizeof(c), 1, f) != 1) return;
}

void TrackObjNameReader(FILE *f, int ind, long p) {
    int c;
    word w[2];
    word parent;

    ASCIIZReader(f, ind, p);
    if (fread(&w, sizeof(w), 1, f) != 1) return;
    if (fread(&parent, sizeof(parent), 1, f) != 1) return;
}

void PivotPointReader(FILE *f, int ind, long p) {
    float pos[3];

    if (fread(&pos, sizeof(pos), 1, f) != 1) return;
}

void SplineFlagsReader(FILE *f, int ind, word flags) {
    int i;
    float dat;

    for (i = 0; i < 16; i++) {
        static const char *flagnames[] = {
            "Tension",
            "Continuity",
            "Bias",
            "Ease To",
            "Ease From",
        };
        if (flags & (1 << i)) {
            if (fread(&dat, sizeof(dat), 1, f) != 1) return;
        }
    }
}

void TrackPosReader(FILE *f, int ind, long p) {
    word n, nf;
    float pos[3];
    word unkown;
    word flags;

    fseek(f, 10, SEEK_CUR);
    if (fread(&n, sizeof(n), 1, f) != 1) return;
    fseek(f, 2, SEEK_CUR);
    while (n-- > 0) {
        if (fread(&nf, sizeof(nf), 1, f) != 1) return;
        if (fread(&unkown, sizeof(unkown), 1, f) != 1) return;
        if (fread(&flags, sizeof(flags), 1, f) != 1) return;
        SplineFlagsReader(f, ind, flags);
        if (fread(&pos, sizeof(pos), 1, f) != 1) return;
    }
}

void TrackRotReader(FILE *f, int ind, long p) {
    word n, nf;
    float pos[4];
    word unkown;
    word flags;

    fseek(f, 10, SEEK_CUR);
    if (fread(&n, sizeof(n), 1, f) != 1) return;
    fseek(f, 2, SEEK_CUR);
    while (n-- > 0) {
        if (fread(&nf, sizeof(nf), 1, f) != 1) return;
        if (fread(&unkown, sizeof(unkown), 1, f) != 1) return;
        if (fread(&flags, sizeof(flags), 1, f) != 1) return;
        SplineFlagsReader(f, ind, flags);
        if (fread(&pos, sizeof(pos), 1, f) != 1) return;
    }
}

void TrackScaleReader(FILE *f, int ind, long p) {
    word n, nf;
    float pos[3];
    word unkown;
    word flags;

    fseek(f, 10, SEEK_CUR);
    if (fread(&n, sizeof(n), 1, f) != 1) return;
    fseek(f, 2, SEEK_CUR);
    while (n-- > 0) {
        if (fread(&nf, sizeof(nf), 1, f) != 1) return;
        if (fread(&unkown, sizeof(unkown), 1, f) != 1) return;
        if (fread(&flags, sizeof(flags), 1, f) != 1) return;
        SplineFlagsReader(f, ind, flags);
        if (fread(&pos, sizeof(pos), 1, f) != 1) return;
    }
}

void ObjNumberReader(FILE *f, int ind, long p) {
    word n;

    if (fread(&n, sizeof(n), 1, f) != 1) return;
}


// ------------------------------------

struct {
    word id;
    const char *name;
    void (*func)(FILE *f, int ind, long p);
} ChunkNames[] = {
    {CHUNK_RGBF,        "RGB float",        RGBFReader},
    {CHUNK_RGBB,        "RGB byte",         RGBBReader},

    {CHUNK_PRJ,         "Project",          NULL},
    {CHUNK_MLI,         "Material Library", NULL},

    {CHUNK_MAIN,        "Main",             NULL},
    {CHUNK_OBJMESH,     "Object Mesh",      NULL},
    {CHUNK_BKGCOLOR,    "Background color", NULL},
    {CHUNK_AMBCOLOR,    "Ambient color",    NULL},
    {CHUNK_OBJBLOCK,    "Object Block",     ObjBlockReader},
    {CHUNK_TRIMESH,     "Tri-Mesh",         NULL},
    {CHUNK_VERTLIST,    "Vertex list",      VertListReader},
    {CHUNK_FACELIST,    "Face list",        FaceListReader},
    {CHUNK_FACEMAT,     "Face material",    FaceMatReader},
    {CHUNK_MAPLIST,     "Mappings list",    MapListReader},
    {CHUNK_SMOOLIST,    "Smoothings",       SmooListReader},
    {CHUNK_TRMATRIX,    "Matrix",           TrMatrixReader},
    {CHUNK_LIGHT,       "Light",            LightReader},
    {CHUNK_SPOTLIGHT,   "Spotlight",        SpotLightReader},
    {CHUNK_CAMERA,      "Camera",           CameraReader},
    {CHUNK_HIERARCHY,   "Hierarchy",        NULL},

    {CHUNK_VIEWPORT,    "Viewport info",    NULL},
    {CHUNK_MATERIAL,    "Material",         NULL},
    {CHUNK_MATNAME,     "Material name",    MatNameReader},
    {CHUNK_AMBIENT,     "Ambient color",    NULL},
    {CHUNK_DIFFUSE,     "Diffuse color",    NULL},
    {CHUNK_SPECULAR,    "Specular color",   NULL},
    {CHUNK_TEXTURE,     "Texture map",      NULL},
    {CHUNK_BUMPMAP,     "Bump map",         NULL},
    {CHUNK_MAPFILE,     "Map filename",     MapFileReader},

    {CHUNK_KEYFRAMER,   "Keyframer data",   NULL},
    {CHUNK_AMBIENTKEY,  "Ambient key",      NULL},
    {CHUNK_TRACKINFO,   "Track info",       NULL},
    {CHUNK_FRAMES,      "Frames",           FramesReader},
    {CHUNK_TRACKOBJNAME,"Track Obj. Name",  TrackObjNameReader},
    {CHUNK_TRACKPIVOT,  "Pivot point",      PivotPointReader},
    {CHUNK_TRACKPOS,    "Position keys",    TrackPosReader},
    {CHUNK_TRACKROTATE, "Rotation keys",    TrackRotReader},
    {CHUNK_TRACKSCALE,  "Scale keys",       TrackScaleReader},
    {CHUNK_OBJNUMBER,   "Object number",    ObjNumberReader},

    {CHUNK_TRACKCAMERA, "Camera track",             NULL},
    {CHUNK_TRACKCAMTGT, "Camera target track",      NULL},
    {CHUNK_TRACKLIGHT,  "Pointlight track",         NULL},
    {CHUNK_TRACKLIGTGT, "Pointlight target track",  NULL},
    {CHUNK_TRACKSPOTL,  "Spotlight track",          NULL},
    {CHUNK_TRACKFOV,    "FOV track",                NULL},
    {CHUNK_TRACKROLL,   "Roll track",               NULL},
};

int FindChunk(word id) {
    int i;
    for (i = 0; i < sizeof(ChunkNames)/sizeof(ChunkNames[0]); i++)
        if (id == ChunkNames[i].id)
            return i;
    return -1;
}

// ------------------------------------


void ChunkReader(FILE *f, int ind, long p) {
    TChunkHeader h;
    int n;
    long pc;

    while (ftell(f) < p) {
        pc = ftell(f);
        if (fread(&h, sizeof(h), 1, f) != 1) return;
        if (h.len == 0) return;
        n = FindChunk(h.id);
        if (n < 0) {
            fseek(f, pc + h.len, SEEK_SET);
        } else {
            pc = pc + h.len;
            if (ChunkNames[n].func != NULL)
                ChunkNames[n].func(f, ind + 2, pc);
            else
                ChunkReader(f, ind + 2, pc);
            fseek(f, pc, SEEK_SET);
        }
        if (ferror(f))
            break;
    }
}


void load_3ds(char *file_name) {
    FILE *f;
    long p;

    avertices=0;
    afaces=0;
    obj_n=0;

    f = fopen(file_name, "rb");
    if (f == NULL) {
        printf("Can't open %s!\n", file_name);
        exit(1);
    }

    fseek(f, 0, SEEK_END);
    p = ftell(f);
    fseek(f, 0, SEEK_SET);
    ChunkReader(f, 0, p);
}

