/* Graphics library for DGS CO'96 using only      */
/* Modes supported are 8,15 and 16 bits per pixel */
/* D  G  S   C  O  M  P  A  N  Y   1  9  9  6     */
/* Code by HERN                                   */
/* Newest version		                  */ 			

#include <dpmi.h>
#include <go32.h>
#include <sys/movedata.h>
#include <sys/segments.h>
#include <sys/nearptr.h>
#include <sys/farptr.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include "font0.fnt"
#include "font1.fnt"
#include "gfx.h"


#define rep_movsl(src, dest, numwords) \
__asm__ __volatile__ ( \
  "rep\n\t" \
  "movsl" \
  : : "S" (src), "D" (dest), "c" (numwords) \
  : "%ecx", "%esi", "%edi" )


unsigned char *_font;

#define MASK_LINEAR(addr) (addr & 0x000FFFFF)
#define RM_OFFSET(addr)   (MASK_LINEAR(addr) & 0xFFFF)
#define RM_SEGMENT(addr)  ((MASK_LINEAR(addr) & 0xFFFF0000) >> 4)
#define RM_TO_LINEAR(addr) (((addr & 0xFFFF0000) >> 12) + (addr & 0xFFFF))

static __dpmi_regs r;
static int size;

typedef struct VESA_INFO         /* VESA information block structure */
{ 
   unsigned char  VESASignature[4]     __attribute__ ((packed));
   unsigned short VESAVersion          __attribute__ ((packed));
   unsigned long  OEMStringPtr         __attribute__ ((packed));
   unsigned char  Capabilities[4]      __attribute__ ((packed));
   unsigned long  VideoModePtr         __attribute__ ((packed)); 
   unsigned short TotalMemory          __attribute__ ((packed)); 
   unsigned short OemSoftwareRev       __attribute__ ((packed)); 
   unsigned long  OemVendorNamePtr     __attribute__ ((packed)); 
   unsigned long  OemProductNamePtr    __attribute__ ((packed)); 
   unsigned long  OemProductRevPtr     __attribute__ ((packed)); 
   unsigned char  Reserved[222]        __attribute__ ((packed)); 
   unsigned char  OemData[256]         __attribute__ ((packed)); 
} VESA_INFO;



typedef struct MODE_INFO         /* VESA information for a specific mode */
{
   unsigned short ModeAttributes       __attribute__ ((packed)); 
   unsigned char  WinAAttributes       __attribute__ ((packed)); 
   unsigned char  WinBAttributes       __attribute__ ((packed)); 
   unsigned short WinGranularity       __attribute__ ((packed)); 
   unsigned short WinSize              __attribute__ ((packed)); 
   unsigned short WinASegment          __attribute__ ((packed)); 
   unsigned short WinBSegment          __attribute__ ((packed)); 
   unsigned long  WinFuncPtr           __attribute__ ((packed)); 
   unsigned short BytesPerScanLine     __attribute__ ((packed)); 
   unsigned short XResolution          __attribute__ ((packed)); 
   unsigned short YResolution          __attribute__ ((packed)); 
   unsigned char  XCharSize            __attribute__ ((packed)); 
   unsigned char  YCharSize            __attribute__ ((packed)); 
   unsigned char  NumberOfPlanes       __attribute__ ((packed)); 
   unsigned char  BitsPerPixel         __attribute__ ((packed)); 
   unsigned char  NumberOfBanks        __attribute__ ((packed)); 
   unsigned char  MemoryModel          __attribute__ ((packed)); 
   unsigned char  BankSize             __attribute__ ((packed)); 
   unsigned char  NumberOfImagePages   __attribute__ ((packed));
   unsigned char  Reserved_page        __attribute__ ((packed)); 
   unsigned char  RedMaskSize          __attribute__ ((packed)); 
   unsigned char  RedMaskPos           __attribute__ ((packed)); 
   unsigned char  GreenMaskSize        __attribute__ ((packed)); 
   unsigned char  GreenMaskPos         __attribute__ ((packed));
   unsigned char  BlueMaskSize         __attribute__ ((packed)); 
   unsigned char  BlueMaskPos          __attribute__ ((packed)); 
   unsigned char  ReservedMaskSize     __attribute__ ((packed)); 
   unsigned char  ReservedMaskPos      __attribute__ ((packed)); 
   unsigned char  DirectColorModeInfo  __attribute__ ((packed));
   unsigned long  PhysBasePtr          __attribute__ ((packed)); 
   unsigned long  OffScreenMemOffset   __attribute__ ((packed)); 
   unsigned short OffScreenMemSize     __attribute__ ((packed)); 
   unsigned char  Reserved[206]        __attribute__ ((packed)); 
} MODE_INFO;



typedef struct PM_INFO           /* VESA 2.0 protected mode interface */
{
   unsigned short setWindow            __attribute__ ((packed)); 
   unsigned short setDisplayStart      __attribute__ ((packed)); 
   unsigned short setPalette           __attribute__ ((packed)); 
   unsigned short IOPrivInfo           __attribute__ ((packed)); 
   /* pmode code is located in this space */
} PM_INFO;

/*
typedef struct MODE_INFO
{
   unsigned char  Reserved0[40];
   unsigned long  PhysBasePtr;
   unsigned char  Reserved1[214];
} MODE_INFO;

static MODE_INFO mode_info;*/

static VESA_INFO vesa_info;            /* SVGA info block */
static MODE_INFO mode_info;            /* info for this video mode */


unsigned long vram;


typedef struct pal8_ex
 {
   unsigned char b;
   unsigned char g;
   unsigned char r;
   unsigned char n;

 } pal8_ex;

float lxmin;
float lxmax;
float lymin;
float lymax;



/**/
int open_lib()
{
  if (__djgpp_nearptr_enable()==0)
    {
      printf("#ERROR: Nearptr enable()\n");
      exit(-1);
    };
}
/**/
int close_lib()
{
  __djgpp_nearptr_disable();
}
/**/
__inline__ void copy(void *src, void *dst, int count)
{
  rep_movsl(src,dst,count);
}
/**/
void set_font(int num)
{
  switch (num)
    {
      case FONT_STANDART: _font=_font0; break;
      case FONT_COMPUTER: _font=_font1; break;
    }
}
/**/
void text_8(char *img,int x,int y,int col,const char *t)
{
  int c;
  int ad,i,j,k,ado,el,nx,mask;
  char b;

  ado=y*screen_w;

  k=0;
  while (*t)
  {
     nx=ado+x;
     el=*t<<3;

     for (i=0;i<=7;i++)
        {
           ad=nx;
           mask=1 << (7-i);

           for (j=el;j<el+8;j++)
              {
               if ((_font[j] & mask)!=0)
                 c=col;
               else
                 c=text_bkg;

               if (c!=-1)

               *(img+ad)=c;

               ad+=screen_w;

              }
           nx++;
        }
     x+=8;
     t++;
  }
}
/**/
void text_15(char *img,int x,int y,int col,const char *t)
{
  int c;
  int ad,i,j,k,ado,el,nx,mask;
  char b;

  ado=y*screen_w;

  k=0;
  while (*t)
  {
     nx=ado+x;
     el=*t<<3;

     for (i=0;i<=7;i++)
        {
           ad=nx;
           mask=1 << (7-i);

           for (j=el;j<el+8;j++)
              {
               if ((_font[j] & mask)!=0)
                 c=col;
               else
                 c=text_bkg;

               if (c!=-1)

               *((short *)img+ad)=c;

               ad+=screen_w;

              }
           nx++;
        }
     x+=8;
     t++;
  }
}
/**/
void textf(char *img,int x,int y,int col,const char *t,...)
{
   char buf[256];

   va_list ap;
   va_start(ap,t);
   vsprintf(buf,t,ap);
   va_end(ap);
   text(img,x,y,col,buf);

}
/**/
void vbl()
{
  while(!(inportb(0x3DA) & 8));
  while(inportb(0x3DA) & 8);
}
/**/
int line_8(char *img,int x1,int y1,int x2,int y2,int c)
{

  int dx,dy,i,adr,x,x0,t,w,h;



  w=screen_w-1;
  h=screen_h-1;

  if (y2<y1)
    {
      t=x2;x2=x1;x1=t;
      t=y2;y2=y1;y1=t;
    }


  dy=y2-y1;

  if (dy==0) return;
  if (y1>h)  return;
  if (y2<0)  return;


  dx=((x2-x1)<<16)/dy;


  x0=x1<<16;

  if (y2>h) y2=h;
  if (y1<0)
    {
      dy-=y1;
      x0+=(-y1)*dx;
      y1=0;
    }

  adr=y1*screen_w;

  dy=y2-y1;

  if (dy>0)
  {

      while (dy--)
    
       {
    
    
         x=x0>>16;
    
         if (x>=0 && x<=w)
         *(img+adr+x)=c;
         adr+=screen_w;
    
         x0+=dx;
    
       }
  }

}
/**/
int line_15(char *img,int x1,int y1,int x2,int y2,int c)
{

  int dx,dy,i,adr,x,x0,t,w,h;



  w=screen_w-1;
  h=screen_h-1;

  if (y2<y1)
    {
      t=x2;x2=x1;x1=t;
      t=y2;y2=y1;y1=t;
    }


  dy=y2-y1;
  if (dy==0) return;
  if (y1>h)  return;
  if (y2<0)  return;


  dx=((x2-x1)<<16)/dy;


  x0=x1<<16;

  if (y2>h) y2=h;
  if (y1<0)
    {
      dy-=y1;
      x0+=(-y1)*dx;
      y1=0;
    }

  adr=y1*screen_w;

  dy=y2-y1;

  if (dy>0)
  {

      while (dy--)
    
       {
    
    
         x=x0>>16;
    
         if (x>=0 && x<=w)
         *((short *)img+adr+x)=c;
         adr+=screen_w;
    
         x0+=dx;
    
       }
  }

}
/**
typedef struct
  {
    int all,l,r,b,t;
  } outcode;

/*
float xmin=50;
float xmax=640-50;
float ymin=50;
float ymax=480-50;



outcode code(float x,float y)
{
  outcode c;


  c.all=0;
  c.l=0;
  c.r=0;
  c.b=0;
  c.t=0;

  if (y>ymax)
    {
      c.t=1;
      c.all+=0x8;
    }
  if (y<ymin)
    {
      c.b=1;
      c.all+=0x4;
    }
  if (x>xmax)
    {
      c.r=1;
      c.all+=0x2;
    }
  if (x<xmin)
    {
      c.l=1;
      c.all+=0x1;
    }
  return c;
}
/**/
void hline(char *img,float x1,float x2,float y,int col)
{
  float x;
  int adr;
  int dx;
  int xp1,xp2,i;

  if (x1>x2)
    {
      x=x1;x1=x2;x2=x;
    }

  if (x2<lxmin) return;
  if (x1>lxmax) return;
  if (y<lymin)  return;
  if (y>lymax)  return;

  if (x1<lxmin) x1=lxmin;
  if (x2>lxmax) x2=lxmax;

  adr=(y*screen_w);
  dx=x2-x1;
  xp1=x1;
  xp2=x2;

/*
  asm ("

         cld;
         rep; stosw

       "
       :
       : "D" (img+adr),"a" (col),"c" (dx)
       : "%edi","%eax","%ecx"
      );*/
  for (i=xp1;i<=xp2;i++) *((short *)img+adr+i)=col;

}
/**/
void vline(char *img,float y1,float y2,float x,int col)
{
  float y;
  int i,adr;

  if (y1>y2)
    {
      y=y1;y1=y2;y2=y;
    }

  if (y2<lymin) return;
  if (y1>lymax) return;
  if (x<lxmin)  return;
  if (x>lxmax)  return;

  if (y1<lymin) y1=lymin;
  if (y2>lymax) y2=lymax;

  adr=(y1*screen_w+x);

  for (i=y1;i<=y2;i++)
    {
      *((short *)img+adr)=col;
      adr+=screen_w;

    }

}
/**/
void draw_line(char *img,int x1, int y1, int x2, int y2, int col)
{
	int sgndistx1, sgndisty1, sgndistx2, sgndisty2;
	int counter2, distx, disty, absdistx, absdisty;
	unsigned int counter;
   int ad,scr_w1,scr_w2;

  	distx = x2 - x1;
  	disty = y2 - y1;


  
  	if (distx < 0)
  		sgndistx1 = -1;
  	else if(distx > 0)
  		sgndistx1 = 1;
  	else
  		sgndistx1 = 0;
  
  	if (disty < 0)
  		sgndisty1 = -1;
  	else if (disty > 0)
  		sgndisty1 = 1;
  	else
  		sgndisty1 = 0;
  
  	absdistx = distx;
  	if (absdistx < 0)
  		absdistx *= -1;
  
  	absdisty = disty;
  	if (absdisty < 0)
  		absdisty *= -1;
  
  	if (absdistx <= absdisty)
  	{
  		sgndistx2 = 0;
  		sgndisty2 = sgndisty1;
  
  		counter2 = absdistx;
  		absdistx = absdisty;
  		absdisty = counter2;
  	}
  	else
  	{
  		sgndistx2 = sgndistx1;
  		sgndisty2 = 0;
  	}
  
  	counter2 = absdistx >> 1;
  
   ad=y1*screen_w;

   scr_w1=screen_w;
   scr_w2=screen_w;
   if (sgndisty1<0) scr_w1=-scr_w1;
   if (sgndisty2<0) scr_w2=-scr_w2;
   if (sgndisty1==0) scr_w1=0;
   if (sgndisty2==0) scr_w2=0;
  
  	for (counter = 0; counter != absdistx; counter++)
  	{

  	   *((short*)img+ad+x1) = col;
  
  		counter2 += absdisty;
  
  		if (counter2 >= absdistx)
  		{
  			 counter2 -= absdistx;
  			 x1 += sgndistx1;
  			 y1 += sgndisty1;
          ad += scr_w1;
  		}
  		else
  		{
  		  x1 += sgndistx2;
  		  y1 += sgndisty2;
        ad += scr_w2;
  		}
  	}

}
typedef struct clip {int l,r,t,b,all;} clip;

/**/
inline clip lcliptest(float x,float y)
{
  clip c;
  c.l=c.r=c.t=c.b=c.all=0;

  if (x<lxmin) c.l=0x1;
  if (x>lxmax) c.r=0x2;
  if (y<lymin) c.t=0x4;
  if (y>lymax) c.b=0x8;

  c.all=c.l+c.r+c.t+c.b;

  return c;

}
/**/
void clip_line(char *img,float x1,float y1,float x2, float y2,int col)
{
  clip       c,c1,c2;
  int        i,j,t;
  float      a,b,x,y,dx,dy;
  float      xp,yp;
  static int xpi[2],ypi[2];
  double     a1;


  dx=x2-x1;
  dy=y2-y1;

  if (dx==0.0 && dy==0.0) return;

  c1=lcliptest(x1,y1);
  c2=lcliptest(x2,y2);

  if (c1.all+c2.all==0 && dx!=0 && dy!=0)
    {
      draw_line(img,x1,y1,x2,y2,col);
      return;
    }

  if (dy==0.0) {hline(img,x1,x2,y1,col);return;}
  if (dx==0.0) {vline(img,y1,y2,x1,col);return;}

  a=dy/dx;
  b=y1-a*x1;
  a1=1/a;

  j=0;
  t=0;

  if (c1.all==0)
    {
      xpi[0]=x1;
      ypi[0]=y1;
      j=1;
      c=c2;
      t=1;
    }
  if (c2.all==0)
    {
      xpi[0]=x2;
      ypi[0]=y2;
      j=1;
      c=c1;
      t=1;
    }



  for (i=j;i<2;i++)
    {
      if (t==0)
        if (i==0)
          c=c1;
        else
          c=c2;


      if (c.t)
        {
          x=(lymin-b)*a1;
          y=lymin;

          if (x>=lxmin && x<=lxmax)
            {
              xp=x;
              yp=y;
            }

        } else

      if (c.b)
        {
          x=(lymax-b)*a1;
          y=lymax;

          if (x>=lxmin && x<=lxmax)
            {
              xp=x;
              yp=y;
            }
        }

      if (c.l)
        {
          x=lxmin;
          y=a*lxmin+b;
          if (y>=lymin && y<=lymax)
            {
              xp=x;
              yp=y;
            }
        } else

      if (c.r)
        {
          x=lxmax;
          y=a*lxmax+b;

          if (y>=lymin && y<=lymax)
            {
              xp=x;
              yp=y;
            }
        }

    
      xpi[i]=x;
      ypi[i]=y;
    }

  draw_line(img,xpi[0],ypi[0],xpi[1],ypi[1],col);



}
/**
int test=1,k;
void line2(char *img,float x1, float y1, float x2, float y2, int col)
{

   float dx,dy,rx,ry;
   float x,y;
   int xp,yp;

   int accept,done;

   float h,w;
   outcode c,c0,c1;

   w=screen_w-1;
   h=screen_h-1;

   dx=x2-x1;
   dy=y2-y1;


   if (test) {

   if (dx==0.0 && dy==0.0) return;

   if (dy==0.0) {hline(img,x1,x2,y1,col);return;}
   if (dx==0.0) {vline(img,y1,y2,x1,col);return;}

   accept=0;
   done=1;
   c0=code(x1,y1);
   c1=code(x2,y2);

   k=0;


   do
     {
        if (c0.all==0 && c1.all==0)
          {
            accept=1;
            done=1;
          }
        else
     
        if (c0.all&c1.all!=0)
          done=1;
        else
          {
              done=0;
            if (c0.all!=0)
              c=c0;
            else
              c=c1;

            dx=x2-x1;
            dy=y2-y1;
            rx=dy/dx;
            ry=dx/dy;

            if (c.t)
              {
                  x=x1+ry*(ymax-y1);
                  y=ymax;
              } else
            if (c.b)
              {
                  x=x1+ry*(ymin-y1);
                  y=ymin;
              } else
            if (c.r)
              {
                  y=y1+rx*(xmax-x1);
                  x=xmax;
              } else
            if (c.l)
              {
                  y=y1+rx*(xmin-x1);
                  x=xmin;
              }
            if (c.all==c0.all)
              {
                x1=x;
                y1=y;
                c0=code(x1,y1);
              } else
              {
                x2=x;
                y2=y;
                c1=code(x2,y2);
              }
            k++;
          }
     }
   while (!done);

   } else accept=1;

//   if (accept) line(img,x1,y1,x2,y2,col);
  
   goo:

   if (accept)
     draw_line(img,x1,y1,x2,y2,col);
}
/**/
void set_procedures(void)
{
  if (screen_bpp==8)
    {
      text=text_8;
//      line=line_8;
    }
  if (screen_bpp==15 || screen_bpp==16)
    {
      text=text_15;
//      line=line_15;
    }
}
/**/
static int get_vesa_info()
{
   int c;

   for (c=0; c<(int)sizeof(VESA_INFO); c++)
      _farpokeb(_dos_ds, MASK_LINEAR(__tb)+c, 0);

   dosmemput("VBE2", 4, MASK_LINEAR(__tb));

   r.x.ax = 0x4F00;
   r.x.di = RM_OFFSET(__tb);
   r.x.es = RM_SEGMENT(__tb);
   __dpmi_int(0x10, &r);
   if (r.h.ah)
      return -1;

   dosmemget(MASK_LINEAR(__tb), sizeof(VESA_INFO), &vesa_info);

   if (strncmp(vesa_info.VESASignature, "VESA", 4) != 0)
      return -1;

   return 0;
}
/**/
static int get_mode_info(int mode)
{
   int c;

   for (c=0; c<(int)sizeof(MODE_INFO); c++)
      _farpokeb(_dos_ds, MASK_LINEAR(__tb)+c, 0);

   r.x.ax = 0x4F01;
   r.x.di = RM_OFFSET(__tb);
   r.x.es = RM_SEGMENT(__tb);
   r.x.cx = mode;
   __dpmi_int(0x10, &r);
   if (r.h.ah)
      return -1;

   dosmemget(MASK_LINEAR(__tb), sizeof(MODE_INFO), &mode_info);
   return 0;
}
/**/
int open_gfx(gfx_mode gmode)
{
  __dpmi_meminfo meminfo;
  unsigned long ul_addr;
  unsigned int size;
  int c,b;

  size=(gmode.size_x * gmode.size_y);
  if (gmode.bpp>=15) size<<=1;

  screen_bpp = gmode.bpp;
  screen_w   = gmode.size_x;
  screen_h   = gmode.size_y;

  lxmin=0.0;
  lxmax=gmode.size_x-1;
  lymin=0;
  lymax=gmode.size_y-1;

  set_procedures();
  set_font(FONT_STANDART);

  r.x.ax = 0x4F01;
  r.x.di = RM_OFFSET(__tb);
  r.x.es = RM_SEGMENT(__tb);
  r.x.cx = 0x4000|gmode.mode;
  __dpmi_int(0x10, &r);
  if (r.h.ah)
    return -1;
  dosmemget(MASK_LINEAR(__tb), sizeof(MODE_INFO), &mode_info);

  meminfo.address = mode_info.PhysBasePtr;
  meminfo.size = size;
  __dpmi_physical_address_mapping(&meminfo);

  vram = meminfo.address;

  __dpmi_lock_linear_region(&meminfo);

  r.x.ax=0x4F02;
  r.x.bx=0x4000|gmode.mode;
  __dpmi_int(0x10,&r);

  __dpmi_get_segment_base_address(_dos_ds, &ul_addr);

  screen=(char *) ul_addr + vram - __djgpp_base_address;

  return 0;
}
/**/
int find_vesa_mode(int w, int h, int color_depth, int vbe2)
{
   #define MAX_VESA_MODES 1024

   unsigned short mode[MAX_VESA_MODES];
   int memorymodel, bitsperpixel;
   int redmasksize, greenmasksize, bluemasksize;
   int greenmaskpos;
   int reservedmasksize, reservedmaskpos;
   int c, modes;
   long mode_ptr;

   if (get_vesa_info() != 0) {
      return -1;
   }

   if ((vesa_info.VESAVersion < 0x200) && (vbe2)) {
      return -2;
   }

   mode_ptr = RM_TO_LINEAR(vesa_info.VideoModePtr);
   modes = 0;

   while ((mode[modes] = _farpeekw(_dos_ds, mode_ptr)) != 0xFFFF) {
      modes++;
      mode_ptr += 2;
   }

   switch (color_depth) {

      case 8:
      default:
	 memorymodel = 4;
	 bitsperpixel = 8;
	 redmasksize = greenmasksize = bluemasksize = 0;
	 greenmaskpos = 0;
	 reservedmasksize = 0;
	 reservedmaskpos = 0;
	 break;

      case 15:
         memorymodel = 6;
         bitsperpixel = 15;
	 redmasksize = greenmasksize = bluemasksize = 5;
	 greenmaskpos = 5;
	 reservedmasksize = 1;
	 reservedmaskpos = 15;
         break;

      case 16:
         memorymodel = 6;
         bitsperpixel = 16;
	 redmasksize = bluemasksize = 5;
	 greenmasksize = 6;
	 greenmaskpos = 5;
	 reservedmasksize = 0;
	 reservedmaskpos = 0;
         break;

      case 24:
         memorymodel = 6;
         bitsperpixel = 24;
	 redmasksize = bluemasksize = greenmasksize = 8;
	 greenmaskpos = 8;
	 reservedmasksize = 0;
	 reservedmaskpos = 0;
         break;

      case 32:
         memorymodel = 6;
         bitsperpixel = 32;
	 redmasksize = greenmasksize = bluemasksize = 8;
	 greenmaskpos = 8;
	 reservedmasksize = 8;
	 reservedmaskpos = 24;
         break;
   }

   #define MEM_MATCH(mem, wanted_mem) \
      ((mem == wanted_mem) || ((mem == 4) && (wanted_mem == 6)))

   #define BPP_MATCH(bpp, wanted_bpp) \
      ((bpp == wanted_bpp) || ((bpp == 16) && (wanted_bpp == 15)))

   #define RES_SIZE_MATCH(size, wanted_size, bpp) \
      ((size == wanted_size) || ((size == 0) && ((bpp == 15) || (bpp == 32))))

   #define RES_POS_MATCH(pos, wanted_pos, bpp) \
      ((pos == wanted_pos) || ((pos == 0) && ((bpp == 15) || (bpp == 32))))

   /* search the list of modes */
   for (c=0; c<modes; c++) {
      if (get_mode_info(mode[c]) == 0) {
	 if (((mode_info.ModeAttributes & 25) == 25) &&
	     (mode_info.XResolution == w) && 
	     (mode_info.YResolution == h) && 
	     (mode_info.NumberOfPlanes == 1) && 
	     MEM_MATCH(mode_info.MemoryModel, memorymodel) &&
	     BPP_MATCH(mode_info.BitsPerPixel, bitsperpixel) &&
	     (mode_info.RedMaskSize == redmasksize) &&
	     (mode_info.GreenMaskSize == greenmasksize) &&
	     (mode_info.BlueMaskSize == bluemasksize) &&
	     (mode_info.GreenMaskPos == greenmaskpos) &&
	     RES_SIZE_MATCH(mode_info.ReservedMaskSize, reservedmasksize, mode_info.BitsPerPixel) &&
	     RES_POS_MATCH(mode_info.ReservedMaskPos, reservedmaskpos, mode_info.BitsPerPixel))
	    /* looks like this will do... */
	    return mode[c];
      } 
   }

   /* try the standard mode numbers */
   if ((w == 640) && (h == 400) && (color_depth == 8))
      c = 0x100;
   else if ((w == 640) && (h == 480) && (color_depth == 8))
      c = 0x101;
   else if ((w == 800) && (h == 600) && (color_depth == 8)) 
      c = 0x103;
   else if ((w == 1024) && (h == 768) && (color_depth == 8))
      c = 0x105;
   else if ((w == 1280) && (h == 1024) && (color_depth == 8))
      c = 0x107;
   else 
      return 0; 
   

   if (get_mode_info(c) == 0)
      return c;

   return 0;
}
/**/
void close_gfx(void)
{
  __dpmi_regs r;
  __dpmi_meminfo meminfo;


  meminfo.address = vram;
  __dpmi_free_physical_address_mapping(&meminfo);

  r.x.ax=0x3;
  __dpmi_int(0x10,&r);

}
/**/
void free_vram(void)
{
  __dpmi_regs r;
  __dpmi_meminfo meminfo;


  meminfo.address = vram;
  __dpmi_free_physical_address_mapping(&meminfo);
}
/**/
void set_color(int n,int r,int g,int b)
{
  outportb(0x3c8,n);
  outportb(0x3c9,r);
  outportb(0x3c9,g);
  outportb(0x3c9,b);
}
/**/
void set_palette(char *pal)
{
  outportb(0x3c8,0);
  outportsb(0x3c9,pal,256*3);
}
/**/
char *load_bmp(pal8 *pal,char *filename)
{
  FILE *f;
  int bv,bh;
  int i;
  int size,buffsize;
  pal8_ex bpal[256];
  int p[256];
  unsigned char n;

  char *buff;
  char *bmp;


  f=fopen(filename,"rb");
  fseek(f,18,0);
  fread(&bv,4,1,f);
  fread(&bh,4,1,f);

  size=bv*bh;
  buffsize=size;

  bmp=(char *) malloc(size);
  if (screen_bpp>8)   size<<=1;
  if (screen_bpp==32) size<<=1;

  buff=(char *) malloc(buffsize);

  fseek(f,54,0);
  fread(bpal,256*4,1,f);
  fseek(f,1078,0);
  fread(bmp,bv*bh,1,f);
  fclose(f);


  if (screen_bpp==8)

    {
      for (i=0;i<256;i++)
        {

          pal[i].r=bpal[i].r>>2;
          pal[i].g=bpal[i].g>>2;
          pal[i].b=bpal[i].b>>2;

        }

      free(buff);

    } else

    {

      if (screen_bpp==15)
        for (i=0;i<256;i++)
          p[i]=COL15(bpal[i].r,bpal[i].g,bpal[i].b);
      if (screen_bpp==16)
        for (i=0;i<256;i++)
          p[i]=COL16(bpal[i].r,bpal[i].g,bpal[i].b);
    
      if (screen_bpp==15 || screen_bpp==16)
      for (i=0;i<size;i++)
    
        {
          n=*(bmp+i);
    
          *( (short *) buff+i ) = p[n];
    
        }
    
      if (screen_bpp==32)

      for (i=0;i<size;i++)
    
        {
          n=*(bmp+i);
    
          *(buff+4*i+0) = bpal[n].b;
          *(buff+4*i+1) = bpal[n].g;
          *(buff+4*i+2) = bpal[n].r;
          *(buff+4*i+3) = 0;
    
        }

      free(bmp);
      return buff;
    }



  return bmp;

}
/**/
struct TGA {

   unsigned char IDLength;
   unsigned char ColorMapType;
   unsigned char ImageType;
   unsigned short int CMapStart;
   unsigned short int CMapLength;
   unsigned char CMapDepth;
   unsigned short int Left;
   unsigned short int Top;
   unsigned short int Width;
   unsigned short int Height;
   unsigned char PixelDepth;
   unsigned char ImageDescriptor;

} tga;

char *load_tga(char *filename)
{
  int i,j,e;
  float dx,dy,px,py;
  char b,c;
  short int w;
  FILE *f;
  char *buff;

  f=fopen(filename,"rb");

  if (f==NULL) return NULL;

  fread(&tga.IDLength,1,1,f);
  fread(&tga.ColorMapType,1,1,f);
  fread(&tga.ImageType,1,1,f);
  fread(&tga.CMapStart,1,2,f);
  fread(&tga.CMapLength,1,2,f);
  fread(&tga.CMapDepth,1,1,f);
  fread(&tga.Left,1,2,f);
  fread(&tga.Top,1,2,f);
  fread(&tga.Width,1,2,f);
  fread(&tga.Height,1,2,f);
  fread(&tga.PixelDepth,1,1,f);
  fread(&tga.ImageDescriptor,1,1,f);

  buff=(char *) malloc(tga.Height*tga.Width*2);

  e=0;
  if (tga.ImageType==2)  fread(buff,tga.Height*tga.Width*2,1,f);

  fclose(f);

  printf(" %dx%dx%d\n",tga.Width,tga.Height,tga.PixelDepth);

  return buff;

}
/**/
void show_bmp(char *img,char *bmp,int w,int h)
{
  int i;

  char *scr;



  scr=img;

  if (h>screen_h) h=screen_h;
  if (w>screen_w) w=screen_w;

  if (screen_bpp>=15)

    {

      for (i=0;i<h;i++)

        memcpy(&scr[i*screen_w*2],&bmp[i*w*2],w*2);

    }
  else
    {
      for (i=0;i<h;i++)

        memcpy(&scr[i*screen_w],&bmp[i*w],w);
    }

}
/**/


