#include "3dtypes.h"
#include "poly.h"

/**/

extern int screen_w;
extern int screen_h;
extern char *scr;
static char *tex;

static short *screen;

static long l,dl;


static long left_x,right_x;
static long left_dx,right_dx;

static long left_l;
static long left_dl;

static long left_h;
static long right_h;
static long height;
static long width;
static long inv_height;
static long prestep;


static t2d *min_v,*max_v,*end_v,*start_v;
static t2d *right_v,*left_v;
static long maxy,miny;


/**/
static void calc_right_deltas()
{
  t2d *v1,*v2;

  v1=right_v;

  if (right_v>start_v)
    v2=right_v-1;
  else
    v2=end_v;

  right_v=v2;

  right_h=ceil(v2->y) - ceil(v1->y);

  if (right_h<=0) return;

  prestep = (ceil(v1->y) << 16) - v1->y;

  if (right_h>1)
    {
      height=v2->y - v1->y;
      right_dx = idiv16( v2->x - v1->x , height);
    }
  else
    {
      inv_height = (0x10000 << 14) / (v2->y - v1->y);
      right_dx = imul14(v2->x - v1->x, inv_height);

    }

  right_x  = v1->x + imul16(prestep,right_dx);

  return;

}
/**/
static void calc_left_deltas()
{

  t2d *v1,*v2;

  v1=left_v;

  if (left_v<end_v)
    v2=left_v+1;
  else
    v2=start_v;

  left_v=v2;


  left_h=ceil(v2->y) - ceil(v1->y);

  if (left_h<=0) return;

  prestep = (ceil(v1->y) << 16) - v1->y;

  if (left_h>1)
    {

      height=v2->y - v1->y;
      left_dx  = idiv16( v2->x - v1->x , height);
      left_dl  = idiv16( v2->l - v1->l , height);
    }
  else
    {
      inv_height = (0x10000 << 14) / (v2->y - v1->y);
      left_dx = imul14(v2->x - v1->x, inv_height);
      left_dl = imul14(v2->l - v1->l, inv_height);
    }
    
  left_x=v1->x  + imul16(prestep,left_dx);
  left_l=v1->l  + imul16(prestep,left_dl);

  return;

}
/**/
void draw_gouraud_poly(t2d *va,int vertices,short *pal)
{

  t2d *v0;

  long x1;

  int i;
  int ad;




  start_v = va;
  min_v   = va;
  max_v   = va;


  miny=va->y;
  maxy=va->y;

  v0      = va+1;

  for (i=1;i<vertices;i++)
    {

      if (v0->y<miny)
        {
          miny=v0->y;
          min_v=v0;
        } else
      if (v0->y>maxy)
        {
  
          maxy=v0->y;
          max_v=v0;
        }
      v0++;
    }

  end_v   = v0-1;
  left_v  = min_v;
  right_v = min_v;




  do {

      if (right_v == max_v)
        return;
      calc_right_deltas();
  } while (right_h<=0);

  do {

      if (left_v == max_v)
        return;
      calc_left_deltas();
  } while (left_h<=0);




  ad=ceil(min_v->y)*screen_w;


  for (;;)
    {

        x1 = ceil(left_x);
        width = ceil(right_x) - x1;


        if (width>0)
          {

            prestep = (x1 << 16) - left_x;
            l = left_l  + imul16(prestep, dl);

            asm (

            "
                 pushl %%ebp;


            n_x:

                 movl %%eax,%%ebp;
                 shrl $16,%%ebp;

                 movw (%%esi,%%ebp,2),%%bx;
                 movw %%bx,(%%edi);

                 addl %%edx,%%eax;

                 addl $2,%%edi;

                 decl %%ecx;

                 jg n_x;

                 popl %%ebp;
            end:

                 "
             :
             : "S" (pal), "D" (scr+(ad+x1)*2), "c" (width), "d" (dl), "a" (l)
             : "%eax", "%ebx", "%ecx", "%edx", "%edi"
            );

          }


        if (--right_h<=0)
        do {
      
            if (right_v == max_v)
              return;
            calc_right_deltas();
      
        } while (right_h<=0);

          else

        right_x+=right_dx;


        if (--left_h<=0)
        do {
      
            if (left_v == max_v)
              return;
            calc_left_deltas();
        } while (left_h<=0);

          else
        {
          left_x+=left_dx;
          left_l+=left_dl;
        }


        ad+=screen_w;



    }


}
/**/
void draw_gouraud_triangle(t2df *va,short *pal)
{
  t2df *v1=va;
  t2df *v2=va+1;
  t2df *v3=va+2;

  static t2d  vx[12];
  static t2df **dst,*t;


  double dy1,dy2;
  double d,id;
  int    i,max;

  double  ddl;



  v1->visible=cliptest(v1);
  v2->visible=cliptest(v2);
  v3->visible=cliptest(v3);

  if ( (v1->visible & v2->visible & v3->visible) != VISIBLE )
    return;



  dy1=v2->y - v3->y;
  dy2=v1->y - v3->y;

  d = (v1->x - v3->x) * dy1 -
      (v2->x - v3->x) * dy2;

  if (d == 0.0) return;

  id = 1.0/d*65536.0;

  ddl = ( (v1->l - v3->l) * dy1 - (v2->l - v3->l) * dy2 ) *id ;


  max=3;

  if ( ( (v1->visible | v2->visible | v3->visible) & 0xff )== VISIBLE )
    {
      fistpL(&vx[0].x,&v1->x);
      fistpL(&vx[1].x,&v2->x);
      fistpL(&vx[2].x,&v3->x);
    }
  else
    {
      max=clip(&dst,va);
      if (max<3) return;

      for (i=0;i<max;i++)
        fistpL(&vx[i].x,&dst[i]->x);
    }

  fistp(&dl,&ddl);

  draw_gouraud_poly(vx,max,pal);


}
/**/

